import os
import sys
import pandas as pd
import numpy as np
import warnings
import random 
import time
from progress.bar import IncrementalBar

class TCRsampler():
  """ 
  Class for sampling CDR3s based on specified v-gene and j-gene usage.

  Attributes
  ----------
  ref_df : pd.DataFrame
    Dataframe of reference CDR$, but contain columns ['v_reps','j_reps','cdr3', 'count','freq']

  ref_dict : dict
    dictionary keyed on tuples that point to dataframe of CDR3s

  Notes
  -----
  The goal of tcrsampler is to allow representative CDR3s to be sampled from a background set. By default, the likelihood that a CDR3 is drawn from the background is proportional to the frequency of that
  CDR3 clone in the original sample. That is, more abundant clones will be sampled more frequently than less abundant clones. 
  """
  def __init__(self):
    self.ref_df = None
    self.ref_dict = None

  def _valid_cdr3(self, cdr3):
    """
    Return True only if CDR3 is comprised of valid upper case amino acid letters.

    Parameters
    ----------
    cdr3 : str
      string repressenting amino acid sequence

    Returns
    -------
    valid : bool

    Examples
    --------
    >>> _valid_cdr3("AAAA")
    True
    >>> _valid_cdr3("AA.A")
    False
    """
    amino_acids = ['A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y']
    valid = np.all([aa in amino_acids for aa in cdr3])
    return valid

  def clean_mixcr(self, filename):
    """
    Parameters
    ----------
    filenname : str
      name of mixcr standard output flat .tsv

    Assigns
    -------
    self.ref_df : pd.DataFrame  
      DataFarme with columns ['v_reps','j_reps','cdr3', 'count']
    """

    bar = IncrementalBar('Clean Mixcr     ', max = 1, suffix='%(percent)d%%')
    df = pd.read_csv(filename, sep = "\t")
    
    if 'subject' in df.columns:
      select_these_columns = ['bestVGene','bestJGene', 'aaSeqCDR3','cloneCount', 'cloneFraction','subject']
    else:
      select_these_columns = ['bestVGene','bestJGene', 'aaSeqCDR3','cloneCount', 'cloneFraction']
    
    rename_these_columns = {'subject':'subject',
                            'bestVGene':'v_reps',
                            'bestJGene':'j_reps', 
                            'aaSeqCDR3':'cdr3',
                            'cloneCount':'count',
                            'cloneFraction':'freq'}

    df.rename(columns =rename_these_columns, inplace = True)
    df['v_reps'] = df['v_reps'].apply(lambda x: f"{x}*01")
    df['j_reps'] = df['j_reps'].apply(lambda x: f"{x}*01")
    
    ind = df['cdr3'].str.startswith('C') & df['cdr3'].str.endswith('F')
    df = df[ind]

    ind = df['cdr3'].apply(lambda x: self._valid_cdr3(cdr3= x))
    df = df[ind]
    df = df[rename_these_columns.values()]
    if 'subject' in df.columns:
      df = df[['v_reps','j_reps','cdr3', 'count','freq','subject']].copy().reset_index(drop = True)
    else:
      df = df[['v_reps','j_reps','cdr3', 'count','freq']].copy().reset_index(drop = True)
    bar.next()
    bar.finish()
    self.ref_df = df

  def build_background(self, df = None, max_rows = 100, stratify_by_subject = False, use_frequency= True, make_singleton = False):
    """
    Parameters
    ----------
    df : pd.DataFrame
      DataFrame with ['v_reps','j_reps','cdr3', 'count','freq'] columns
    max_rows : int
      Maximum clones per v,j pair (per subject)  
    stratify_by_subject : bool
      If True, max_rows will apply to v,j,subject. If False, max_rows applies to v,j
    use_frequency : bool
      If True, uses frequency for ranking rows. If False, uses raw counts. 
    make_singleton : bool
      If True, background is still sorted by frequency or counts, but final fequency and counts values are overridden
      and set to 1. 
    Assigns
    -------
    self.ref_dict : dict
      dictionary keyed on (v,j) tuples pointing to pd.DataFrame
    """
    if df is None:
      df = self.ref_df

    if use_frequency:
      col = 'freq'
    else:
      col = 'count'

    dfg =df.groupby(['v_reps','j_reps'])
    
    bar = IncrementalBar('Build Background', max = dfg.ngroups, suffix='%(percent)d%%')
    
    d = dict()
    
    for i,group in dfg:
    
      bar.next()
      # CASE:  stratify_by_subject = False
      if not stratify_by_subject:
        if group.shape[0] > 0:
          if group.shape[0] > max_rows:
            n = max_rows
          else:
            n = group.shape[0]
          d[i] =group.sort_values([col], ascending = False).reset_index(drop = True).iloc[0:n,].copy()
      else: # CASE:  stratify_by_subject = True, another nested groupby, and max_rows applies to subject
        dfgg = group.groupby(['subject'])
        for ii , ggroup in dfgg:
          if ggroup.shape[0] > 0:
            if ggroup.shape[0] > max_rows:
              n = max_rows
            else:
              n = ggroup.shape[0]
            if i not in d.keys():
              d[i] = ggroup.sort_values([col], ascending = False).reset_index(drop = True).iloc[0:n,].copy()
            else: 
              dii = ggroup.sort_values([col], ascending = False).reset_index(drop = True).iloc[0:n,].copy()
              d[i] = pd.concat([d[i],dii])
    bar.finish()
    
    if make_singleton:
      for k,v in d.items():
        v['count'] = 1
        v['freq']  = 1

    self.ref_dict = d

  def sample_background(self,v,j,n=1, d= None, depth = 1, seed =1, use_frequency= True ):
    """
    Parameters
    ----------
    v : str
      v-gene name e.g., 'TRBV10-1*01'
    j : str
      j-gene name e.g., 'TRBJ1-1*01'
    n : int
      number of cdr3 samples to draw for given v,j
    d : dict 
      Dictionary for sampling, generated in by .build_background
    depth : 

    seed : int
      random number generating seed
    use_frequency : bool
      If True, uses frequency for sampling proportionaly. If False, uses raw counts. 

    Returns
    -------
    r: list

    Example 
    -------
    >>> sample_background(v ='TRBV10-1*01', j ='TRBJ1-1*01',n=1, d= None, depth = 1, seed =1, use_frequency= True )
    """
    if d is None:
      d = self.ref_dict


    if use_frequency:
      col = 'freq'
    else:
      col = 'count'
    
    assert isinstance(v, str)
    assert isinstance(j, str)
    assert isinstance(d, dict)
    assert isinstance(depth, int)
    assert isinstance(seed, int)

    try:
      subdf = d[(v,j)]
  
      selection_probability = \
        subdf[ col ] / np.sum(subdf[ col ])

      np.random.seed(seed) 
      
      probabalistic_selection_index = \
        np.random.choice( range(subdf.shape[0]),
        size = n * depth,
        p=selection_probability)

      r = subdf.iloc[probabalistic_selection_index,]['cdr3'].to_list()
      return r

    except KeyError:
      warnings.warn(f"({v},{j} gene usage not available")
      r = [None]
      return r


  def sample(self, v_j_usage, depth = 1, seed = 1, flatten = False, use_frequency= True):
    """
    Sample a reference dictionary based on v and j gene usage 

    Parameters
    ----------
    v_j_usage : list of lists or list of tuples (v-gene. j-gene, n)
      e.g., ['TRBV9*01','TRBJ2-7*01', 2],['TRBV7-7*01', 'TRBJ2-4*01', 4] 
    depth : int
      mulitple of the number of times to sample for a given frequncy. 
    seed : int
      random number initialization
    flatten : bool
      if true return a single list, if false a list of lists 
    
    Returns 
    -------
    result : list
      list of lists if flatten is False, list if flatten is True 

    Example
    -------
    >>> t.sample([['TRBV9*01','TRBJ2-7*01', 2],['TRBV7-7*01', 'TRBJ2-4*01', 4] ])
    """
    assert isinstance(v_j_usage, list)
    assert isinstance(depth, int)
    assert isinstance(seed, int)

    result = list()
    for v,j,n in v_j_usage:
      r = self.sample_background(v = v, j = j, n = n, depth = depth, seed = seed, use_frequency= use_frequency)
      result.append(r)

    if flatten:
      result = list(np.concatenate(result))

    return result




import pytest
from tcrsampler import setup_db

def test_install_all_next_gen(dry_run = True):
	""" with dry_run set to true, test setup_db routine runs without error"""
	setup_db.install_all_next_gen(dry_run = dry_run)

def test_install_nextgen_data_to_db_KeyError_download_from(download_file = "new_nextgen_chains_mouse_A.tsv", 
									download_from = "fhcrc", 
									dry_run = True):
	""" with dry_run set to true, test setup_db routine runs without error"""
	with pytest.raises(KeyError):
		setup_db.install_nextgen_data_to_db(download_file = download_file, 
											download_from = download_from , 
											dry_run = dry_run)

def test_install_nextgen_data_to_db_KeyError_filename(download_file = "new_nextgen_chains_hamster_A.tsv", 
									download_from = "fhcrc", 
									dry_run = True):
	""" with dry_run set to true, test setup_db routine runs without error"""
	with pytest.raises(KeyError):
		setup_db.install_nextgen_data_to_db(download_file = download_file, 
											download_from = download_from , 
											dry_run = dry_run)

def test_install_nextgen_data_to_db_TypeError(download_file = ["new_nextgen_chains_mouse_A.tsv"], 
									download_from = "dropbox", 
									dry_run = True):
	""" with dry_run set to true, test setup_db routine runs without error"""
	with pytest.raises(TypeError) as e:
		setup_db.install_nextgen_data_to_db(download_file = download_file, 
										download_from = download_from , 
										dry_run = dry_run)
		assert e.type is TypeError
		assert e.value.args[0].startswith('The <download_from> arg must be a string')





import pytest 
import os
import pandas as pd
from tcrsampler.sampler import TCRsampler


def test_TCRsampler_init():
	t = TCRsampler()

def test_TCRsampler_clean_mixcr():
	t = TCRsampler()
	fn= os.path.join('tcrsampler' ,'tests', 'pmbc_mixcr_example_data.txt')
	t.clean_mixcr(filename = fn)
	assert isinstance(t.ref_df, pd.DataFrame)

def test_TCRsampler_build():
	t = TCRsampler()
	fn= os.path.join('tcrsampler' ,'tests', 'pmbc_mixcr_example_data.txt')
	t.clean_mixcr(filename = fn)
	t.build_background()
	assert isinstance(t.ref_dict, dict)
	assert isinstance(t.ref_dict.popitem()[1], pd.DataFrame)

def test_TCRsampler_build_singleton():
	t = TCRsampler()
	fn= os.path.join('tcrsampler' ,'tests', 'pmbc_mixcr_example_data.txt')
	t.clean_mixcr(filename = fn)
	t.build_background(stratify_by_subject = True, make_singleton = True)
	r = t.sample_background('TRBV9*01','TRBJ2-7*01', n =10 )


def test_TCRsampler_build_stratified():
    t = TCRsampler()
    fn= os.path.join('tcrsampler' ,'tests', 'pmbc_mixcr_example_data.txt')
    t.clean_mixcr(filename = fn)
    t.build_background(stratify_by_subject = True)
    r = t.sample_background('TRBV9*01','TRBJ2-7*01', n =10 )


def test_prob_sampler_sample_background():
	t = TCRsampler()
	fn= os.path.join('tcrsampler' ,'tests', 'pmbc_mixcr_example_data.txt')
	t.clean_mixcr(filename = fn)
	t.build_background()
	r = t.sample_background('TRBV9*01','TRBJ2-7*01', n =10 )
	assert r == ['CASSRTGSLADEQYF',
				 'CASSATGVVSAQYF',
				 'CASSAWGQVYEQYF',
				 'CASSVSGSPYEQYF',
				 'CASSAWGQVYEQYF',
				 'CASSAWGQVYEQYF',
				 'CASRWGEQYF',
				 'CASSGDDWEQYF',
				 'CASSATGTSGPYEQYF',
				 'CASSSRTSGSNSEQYF']

def test_prob_sampler_sample():
	t = TCRsampler()
	fn= os.path.join('tcrsampler' ,'tests', 'pmbc_mixcr_example_data.txt')
	t.clean_mixcr(filename = fn)
	t.build_background()
	r = t.sample([['TRBV9*01','TRBJ2-7*01', 2]])
	assert r == [['CASSRTGSLADEQYF', 'CASSATGVVSAQYF']]
	r = t.sample([['TRBV9*01','TRBJ2-7*01', 2]], flatten = True)
	assert r == ['CASSRTGSLADEQYF', 'CASSATGVVSAQYF']
	r = t.sample([['TRBV9*01','TRBJ2-7*01', 2],['TRBV7-7*01', 'TRBJ2-4*01', 4] ])
	assert r == [['CASSRTGSLADEQYF', 'CASSATGVVSAQYF'],
 					['CASSLGQAARGIQYF', 'CASSLGQAARGIQYF', 'CASSLGQAARGIQYF', 'CASSLGQAARGIQYF']]


def test_prob_sampler_sample_key_warn():
	t = TCRsampler()
	fn= os.path.join('tcrsampler' ,'tests', 'pmbc_mixcr_example_data.txt')
	t.clean_mixcr(filename = fn)
	t.build_background()
	with pytest.warns(None):
		r = t.sample([['TRBV999*01','TRBJ2-7*01', 2]])
	assert r == [[None]]





import os
import sys

def install_nextgen_data_to_db(download_file, download_from = "dropbox", dry_run = False):
    """
    Function installs next-gen files

    Parameters
    ----------
    download_file : string
        "new_nextgen_chains_mouse_A.tsv"
    download_from : string 
        'dropbox' is only currently available option
    dry_run : bool 
        if True, download commands are printed but no executed
    Returns
    -------
    curl_url_cmd : string
        the command for installing to tcrdist/db/alphabeta_db.tsv_files/*

    """
    if not isinstance(download_from, str):
        raise TypeError('The <download_from> arg must be a string')
    
    download_sources = ["dropbox"]
    if download_from not in download_sources:
        raise KeyError(f"the <download_from> arg must be be one of the following {download_sources}")

    if download_from is "dropbox":
        address = { "britanova_chord_blood.csv" : 'https://www.dropbox.com/s/rkbce72njcei4y8/britanova_chord_blood.csv?dl=1',
                    "emerson_cmv_negative.csv"  : 'https://www.dropbox.com/s/yrozbowxtqumjfl/emerson_cmv_negative.csv?dl=1',
                    "new_nextgen_chains_mouse_A.tsv" : 'https://www.dropbox.com/s/pkpr6p97eworn3q/new_nextgen_chains_mouse_A.tsv?dl=1',
                    "new_nextgen_chains_mouse_B.tsv" : 'https://www.dropbox.com/s/sxgvrj25mnzr20s/new_nextgen_chains_mouse_B.tsv?dl=1',
                    "new_nextgen_chains_human_A.tsv" : 'https://www.dropbox.com/s/9p43c7tscf46dat/new_nextgen_chains_human_A.tsv?dl=1',
                    "new_nextgen_chains_human_B.tsv" : 'https://www.dropbox.com/s/83xbk7jp4xd3pr0/new_nextgen_chains_human_B.tsv?dl=1',
                    "new_nextgen_chains_human_G.tsv" : 'https://www.dropbox.com/s/41w8yl38nr4ey32/new_nextgen_chains_human_A.tsv?dl=1',
                    "new_nextgen_chains_human_D.tsv" : 'https://www.dropbox.com/s/8ysciqrcywdsryp/new_nextgen_chains_human_B.tsv?dl=1'}

        if download_file not in address.keys():
            raise KeyError("download_file must be in {}".format(" ".join(map(str,address.keys))))

    # Where the file is to be installed
    path_file = os.path.realpath(__file__)
    path = os.path.dirname(path_file)
    install_path = os.path.join(path, "db", download_file) #<----- ALPHA BETA destination

    def generate_curl(filename, download_link):
        return('curl -o {} {} -L'.format(filename, download_link))

    curl_url_cmd = generate_curl(install_path, address[download_file])
    sys.stdout.write("RUNNING: {}\n".format(curl_url_cmd) )
    
    if dry_run is False:
        os.system(curl_url_cmd)
    return(curl_url_cmd)


def install_all_next_gen(dry_run = False):
    all_files = ["new_nextgen_chains_mouse_A.tsv",
                "new_nextgen_chains_mouse_B.tsv" ,
                "new_nextgen_chains_human_A.tsv",
                "new_nextgen_chains_human_B.tsv",
                "new_nextgen_chains_human_G.tsv", 
                "new_nextgen_chains_human_D.tsv",
                "britanova_chord_blood.csv",
                "emerson_cmv_negative.csv" ]
    select_files = ["britanova_chord_blood.csv","emerson_cmv_negative.csv" ]

    for fn in select_files:
        install_nextgen_data_to_db(download_file = fn, download_from = 'dropbox', dry_run = dry_run )



from setuptools import setup, find_packages
PACKAGES = find_packages()

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

opts = dict(name='tcrsampler',
            maintainer='Koshlan Mayer-Blackwell',
            maintainer_email='kmayerbl@fredhutch.org',
            description='Sample TCRs according to VDJ gene usage frequency',
            long_description=long_description,
            long_description_content_type='text/markdown',
            url='https://github.com/kmayerb/tcrsampler',
            license='MIT',
            author='Koshlan Mayer-Blackwell',
            author_email='kmayerbl@fredhutch.org',
            version='0.1.3',
            packages=PACKAGES,
            include_package_data=True)

install_reqs = [
      'numpy>=1.18.1',
      'pandas>=0.24.2', 
      'progress>=1.5']

if __name__ == "__main__":
      setup(**opts, install_requires=install_reqs)
